import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:flutterx_application/src/localization/labels.dart';
import 'package:flutterx_application/src/platform/platform_io.dart'
    if (dart.library.html) 'package:flutterx_application/src/platform/platform_html.dart';
import 'package:flutterx_application/src/routing/app_route.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';
import 'package:flutterx_utils/flutterx_utils.dart';

part 'application_binding.dart';
part 'application_impl.dart';

/// [runApp] wrapping the widget tree inside [Application] implementation
///
/// [name] of the application
/// [routes] all routes of the application
/// [initialize] initialize app and return first route. Default returns first route
/// [labels] set of supported [LabelInterface] used inside the application
/// [onLabel] callback called when application [Locale] changes passing the corresponding [LabelInterface].
///           Use this callback to update the labels global variable
/// [materialAppFactory] check [MaterialAppFactory].
///                      Default value uses the default theme so you may want to customize it
Future<void> runApplication({
  required String name,
  required Set<AppRoute> routes,
  InitializeApp? initialize,
  OnUnknownRoute onUnknownRoute = _defaultOnUnknownRoute,
  Set<LabelInterface>? labels,
  ValueSetter<LabelInterface>? onLabel,
  MaterialAppFactory materialAppFactory = _defaultMaterialAppFactory,
}) async {
  WidgetsFlutterBinding.ensureInitialized();
  final initialRoute = initialize == null ? routes.first : await initialize();
  runApp(_ApplicationImpl(
      name: name,
      routes: routes,
      initialRoute: initialRoute.location,
      onUnknownRoute: onUnknownRoute,
      labels: labels,
      onLabel: onLabel,
      materialAppFactory: materialAppFactory));
}

/// Unique representation of your application
///
/// To use this class just call [runApplication] in your main() method instead of [runApp]
///
/// [of] use this to get [Application] instance from everywhere in your app
/// [isInputMethodVisible] [LiveData] that represents true if input method is visible
/// [frameCallback] [LiveEvent] you can use to listen [WidgetsBinding.addPersistentFrameCallback]
/// [supportedLocales] list of [Locale] supported by this application
/// [defaultLocale] is default locale resolved from supported locales
/// [enforcedLocale] can be used to force Application locale regardless of system settings, or null for default
/// [restart] (soft) recreate widget tree containing the whole application. This will not kill application process
/// [exit] close application. (this can kill application process)
/// [addLifecycleObserver] add [LifecycleObserver] and immediately dispatch current [AppLifecycleState] to it
/// [removeLifecycleObserver] remove [LifecycleObserver]
abstract class Application {
  static Application of(BuildContext context) =>
      requireNotNull(context.findRootAncestorStateOfType<_ApplicationImplState>(),
          message: () => 'Application not reachable from current context. Have you called runMaterialApp()?');

  LiveData<bool> get isInputMethodVisible;

  LiveData<AppLifecycleState> get lifecycle;

  List<Locale> get supportedLocales;

  Locale get defaultLocale;

  Locale? enforcedLocale;

  void restart();

  void exit();

  /// Compute default [SystemUiOverlayStyle] from [context]
  static SystemUiOverlayStyle defaultSystemUiStyle(BuildContext context) {
    final theme = Theme.of(context);
    final appBarTheme = theme.appBarTheme;
    final statusBarBrightness = appBarTheme.systemOverlayStyle?.statusBarBrightness ??
        appBarTheme.backgroundColor?.brightness ??
        theme.primaryColorBrightness;
    final navBarColor = theme.scaffoldBackgroundColor.darken(.05);
    return SystemUiOverlayStyle(
        systemNavigationBarColor: navBarColor,
        systemNavigationBarIconBrightness: navBarColor.brightness.negative,
        statusBarColor: Colors.black26,
        statusBarBrightness: statusBarBrightness,
        statusBarIconBrightness:
            appBarTheme.systemOverlayStyle?.statusBarIconBrightness ?? statusBarBrightness.negative);
  }
}

/// Create [SystemUiOverlayStyle] from context.
/// This is used by [AppRoute], default value is [Application.defaultSystemUiStyle]
typedef SystemUiStyleBuilder = SystemUiOverlayStyle Function(BuildContext context);

/// Initialize application
typedef InitializeApp = FutureOr<ActivityRoute> Function();

/// Fallback route
typedef OnUnknownRoute = Future<AppRoute> Function(RouteInformation error);

/// Factory of [MaterialApp] used as parameter of [runApplication]
///
/// Every parameter passed by this closure must be used as parameter of the returned [MaterialApp]
typedef MaterialAppFactory = Widget Function(
  GlobalKey<State<MaterialApp>> key,
  AppRouterDelegate routerDelegate,
  AppRouteInformationProvider routeInformationProvider,
  AppRouteInformationParser routeInformationParser,
  String title,
  Locale? locale,
  Iterable<LocalizationsDelegate> localizationsDelegates,
  Iterable<Locale> supportedLocales,
);

MaterialApp _defaultMaterialAppFactory(
  GlobalKey key,
  AppRouterDelegate routerDelegate,
  AppRouteInformationProvider routeInformationProvider,
  AppRouteInformationParser routeInformationParser,
  String title,
  Locale? locale,
  Iterable<LocalizationsDelegate> localizationsDelegates,
  Iterable<Locale> supportedLocales,
) =>
    MaterialApp.router(
        key: key,
        routerDelegate: routerDelegate,
        routeInformationProvider: routeInformationProvider,
        routeInformationParser: routeInformationParser,
        title: title,
        locale: locale,
        localizationsDelegates: localizationsDelegates,
        supportedLocales: supportedLocales,
        debugShowCheckedModeBanner: false);

Future<AppRoute> _defaultOnUnknownRoute(RouteInformation error) => SynchronousFuture(
    ActivityRoute.builder(name: '404', builder: (context, args) => const Scaffold(body: Center(child: Text('404')))));
