part of 'app_route.dart';

// ignore: prefer_mixin
class AppRouterDelegate extends RouterDelegate<AppPage> with ChangeNotifier, PopNavigatorRouterDelegateMixin {
  late AppRouteInformationProvider provider;
  late AppRouteInformationParser parser;

  final Map<String, AppRoute> _routes;
  final String _initialLocation;
  final OnUnknownRoute _onUnknownRoute;
  final List<AppPage> _stack = [];

  @override
  final GlobalKey<NavigatorState> navigatorKey = GlobalKey();

  @override
  AppPage? get currentConfiguration => _stack.lastOrNull;

  RouteInformation get _entrypoint {
    var location = WidgetsBinding.instance?.window.defaultRouteName ?? Navigator.defaultRouteName;
    if (location == Navigator.defaultRouteName) location = _initialLocation;
    return RouteInformation(location: location);
  }

  AppRouterDelegate(Iterable<AppRoute> routes, this._initialLocation, this._onUnknownRoute)
      : _routes = Map.fromEntries(routes.map((route) => MapEntry(route.location, route))) {
    provider = AppRouteInformationProvider._(this);
    parser = AppRouteInformationParser._(this);
  }

  @override
  Future<void> setInitialRoutePath(AppPage configuration) {
    final stack = parseRouteStack();
    if (stack == null) return super.setInitialRoutePath(configuration);
    for (final location in stack) {
      final page = _parseRouteInformationInternal(location, const {});
      if (page != null) _addRoute(page);
    }
    return SynchronousFuture(null);
  }

  @override
  Future<void> setNewRoutePath(AppPage configuration) {
    _addRoute(configuration);
    return SynchronousFuture(null);
  }

  @override
  Widget build(BuildContext context) => Navigator(
      key: navigatorKey, pages: _stack.toList(growable: false), onPopPage: _onPopPage, restorationScopeId: 'app_nav');

  bool _onPopPage(Route route, dynamic result) {
    final appPage = route.settings;
    if (route.didPop(result) && appPage is AppPage) {
      _removeRoute(appPage);
      return true;
    }
    return false;
  }

  void _addRoute(AppPage page) {
    _stack.add(page);
    notifyListeners();
  }

  void _replaceRoute(AppPage page) {
    _stack.removeLast();
    _stack.add(page);
    notifyListeners();
  }

  void _removeRoute(AppPage page) {
    _stack.remove(page);
    notifyListeners();
  }

  Future<AppPage> _parseRouteInformation(RouteInformation routeInformation) {
    final page = _parseRouteInformationInternal(
        routeInformation.location ?? Navigator.defaultRouteName, routeInformation.state as RouteArgs? ?? const {});
    return page == null
        ? _onUnknownRoute(routeInformation).then((value) => value.createPage())
        : SynchronousFuture(page);
  }

  AppPage? _parseRouteInformationInternal(String location, RouteArgs defaultArgs) {
    final route = _routes[location];
    if (route == null) return null;
    final args = parseRouteInformationState(location) ?? defaultArgs;
    return route.createPage(args: args);
  }

  RouteInformation _restoreRouteInformation(AppPage configuration) {
    final location = configuration.route.location;
    restoreRouteInformationState(location, configuration.arguments);
    restoreRouteStack(_stack.map((page) => page.route.location).toList(growable: false));
    return RouteInformation(location: location, state: configuration.arguments);
  }

  static AppRouterDelegate of(BuildContext context) => Router.of(context).routerDelegate as AppRouterDelegate;
}
